/*
   Common Routines
   --------------------------------------------------------------------
   Protolog - The Internet Protocols logger

   Copyright (C) 1998 Diego Javier Grigna <diego@grigna.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"

/* Version of the Protolog package */
char *versionstr="1.0.0";

/* If set Protolog won't resolve domain names */
int flag_dont_resolve;

/*
 * Initialize signals
 */
void init_signals( void)
{
 signal( SIGINT , SIG_IGN);
 signal( SIGHUP , SIG_IGN);
 signal( SIGTERM, SIG_IGN);
 signal( SIGQUIT, SIG_IGN);
 signal( SIGKILL, SIG_IGN);
 signal( SIGPIPE, SIG_IGN);
 signal( SIGSTOP, SIG_IGN);
 signal( SIGTSTP, SIG_IGN);
 signal( SIGTTIN, SIG_IGN);
 signal( SIGTTOU, SIG_IGN);
}

void gobackground( void)
{
 int pid, fd;

 if( (pid = fork()))
     exit(0);
 else 
     if( pid == -1) {
         fprintf( stderr, "\nFatal error: %s\n", strerror( errno));
         exit( -1);
     }

 fd = open( "/dev/tty", O_RDWR);
 ioctl( fd, TIOCNOTTY, 0);
 close( fd);

 close( 0);
 close( 1);
 close( 2);

 setsid();
 chdir( "/");
}

char *resolve_host_name( unsigned long int addr)
{
 static char buf[ 4096];
 struct in_addr ina;
 struct hostent *he;

 if( flag_dont_resolve) {
     strcpy( buf, "not resolving");
     return buf;
 }

 ina.s_addr = addr;

 he = gethostbyaddr( (char *) &ina, sizeof( struct in_addr), AF_INET);

 if( he == NULL)
     strcpy( buf, inet_ntoa( ina));
 else
     strcpy( buf, he->h_name);

 return buf;
}

char *get_serv_name( unsigned short port)
{
 struct servent *se;
 static char buf[ 2048];

 se = getservbyport( port, "tcp");

 if( se == NULL)
     sprintf( buf, "%d", ntohs( port));
 else
     sprintf( buf, "%s", se->s_name);

 return buf;
}

/*
 * Run only if we are root.
 */

void check_uid( char *progname)
{
 if( geteuid() || getuid()) {
     fprintf( stderr,"%s: Only root can run this program!\n", progname);
     exit( -1);
 }
}

char *replicate( char c, int size)
{
 static char rstr[ 2048];

 memset( &rstr, c, size);

 return rstr;
}

