/*
   Ipretplog - Protolog RAW packet format interpreter
   --------------------------------------------------------------------
   Protolog - The Internet Protocols logger

   Copyright (C) 1998 Diego Javier Grigna <diego@grigna.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"

/* Function prototypes */

void mainloop( FILE *fp);
void print_header_ip( struct iphdr iph);
void print_header_ipoptions( char *buf, int size);
void print_header_tcp( struct tcphdr tcph);
void print_header_udp( struct udphdr udph);
void print_header_icmp( struct icmphdr ich);
void print_data( char *buf, char *pkttype, int size);
void print_ala_hexdump( char *buf, int size);
void print_hex( char *buf, int size);
void print_ascii( char *buf, int size);


int main( int argc, char *argv[])
{
 FILE *fp;
 struct stat stbuf;
 int ch;

 flag_dont_resolve = 1;

 if( argc < 2) {
     mainloop( stdin);
     return 0;
 }

 while(( ch = getopt( argc, argv, "r")) != EOF){
         if( ch == 'r')
             flag_dont_resolve = 0;
 }

 if( optind >= argc)
     fp = stdin;
 else {
     fp = fopen( argv[ optind], "r");

     if( fp == NULL) {
         fprintf( stderr, "%s: %s : %s\n", argv[ 0], argv[ optind], strerror( errno));
         return -1;
     }

     fstat( fileno( fp), &stbuf);

     if( ( stbuf.st_mode & S_IFMT) == S_IFDIR){
           fprintf( stderr, "\"%s\" is a directory\n", argv[ optind]);
           return -1;
     }

     if( ( stbuf.st_mode & S_IFMT) != S_IFREG){
           fprintf( stderr, "\"%s\" is not a regular file\n", argv[ optind]);
           return -1;
     }

 }

 mainloop( fp);

 return 0;
}

void mainloop( FILE *fp)
{
 struct iphdr   iph; /* IP header                       */
 struct tcphdr tcph; /* TCP header                      */
 struct udphdr udph; /* UDP header                      */
 struct icmphdr ich; /* ICMP header                     */
 char buf[ 65535];   /* Buffer to read packets          */
 time_t t;           /* Time when the packet was logged */
 int iphlen;         /* IP header length                */
 int hsize;          /* Header size (TCP, UDP or ICMP   */
 int totlen;         /* Total length size               */
 int datas;          /* Data size                       */
 int packet;         /* Number of packets               */
 char guion[ 128];
 char *pkttype;

 iphlen = hsize = totlen = datas = packet = 0;

 sprintf( guion, replicate( '-', 80));

 while( !feof( fp)) {
        if( fread( &t, sizeof( t), 1, fp) <= 0) 
            break;

        packet++;

        printf( "%-22.22s Begin of packet number: %10d %-22.22s\n"
                  , guion
                  , packet
                  , guion);

        printf( "%-20.20s arrival date: %-24.24s %-20.20s\n"
                  , guion
                  , ctime( &t)
                  , guion);

        if( fread( &iph, sizeof( iph), 1, fp) <= 0)
            break;

        print_header_ip( iph);

        iphlen  = iph.ihl << 2;
        iphlen -=20;

        if( iphlen > 0) {
            if( fread( buf, iphlen, 1, fp) <= 0)
                break;
            print_header_ipoptions( buf, iphlen);
        }

        if( iph.protocol == IPPROTO_TCP) {
            if( fread( &tcph, sizeof( tcph), 1, fp) <= 0)
                break;
            hsize = sizeof( tcph);
            pkttype = "TCP";
            print_header_tcp( tcph);
        } else 
            if( iph.protocol == IPPROTO_UDP) {
                if( fread( &udph, sizeof( udph), 1, fp) <= 0)
                    break;
                hsize = sizeof( udph);
                pkttype = "UDP";
                print_header_udp( udph);
            } else
                if( iph.protocol == IPPROTO_ICMP) {
                    if( fread( &ich, sizeof( ich), 1, fp) <= 0)
                        break;
                    hsize = sizeof( ich);
                    pkttype = "ICMP";
                    print_header_icmp( ich);
                } else {
                    printf( "Protocol not supported by Protolog\n");
                    continue;
                }

        totlen = sizeof( iph) + hsize + iphlen;

        if( totlen < htons( iph.tot_len )) {
            datas = htons( iph.tot_len) - totlen;
            if( fread( buf, datas, 1, fp) <= 0)
                break;
            print_data( buf, pkttype, datas);
        }

        printf( "%-22.22s End of packet number:   %10d %-22.22s\n\n"
                  , guion
                  , packet
                  , guion);

 }

}

void print_header_ip( struct iphdr iph)
{
 struct protoent *proto;
 struct in_addr addrs;

 printf( "******* IP Header *******\n");
 printf( "IP Version                  : %-5d (IPv%d)\n"     , iph.version        , iph.version        );
 printf( "Header Length               : %-5d (%d)\n"        , iph.ihl << 2       , iph.ihl            );
 printf( "Type Of Service             : %-5d (0x%x)\n"      , iph.tos            , iph.tos            );
 printf( "Total lenght of packet      : %-5d (0x%x) bytes\n", htons( iph.tot_len), htons( iph.tot_len));
 printf( "Identification              : %-5d (0x%x)\n"      , htons( iph.id)     , htons( iph.id)     );
 printf( "3 bit flags                 : %-5d (0x%x)\n"      , (( iph.frag_off) & 0xe000) >> 13, ((iph.frag_off) & 0xe000) >> 13);
 printf( "13 bit Fragmentation offset : %-5d (0x%x)\n"      , ( iph.frag_off) & 0x1fff,( iph.frag_off) & 0x1fff);
 printf( "Time To Live                : %-5d (0x%x)\n"      , iph.ttl            , iph.ttl            );

 proto = NULL;
 proto = getprotobynumber( iph.protocol);

 if( proto == NULL)
     printf( "Protocol                    : Unknow protocol! %-5d (0x%x)\n", iph.protocol, iph.protocol);
 else
     printf( "Protocol                    : %s\n", proto->p_name);

 printf( "CheckSum                    : %-5d (0x%x)\n"      , htons( iph.check), htons( iph.check));

 addrs.s_addr = iph.saddr;
 printf( "IP Source address (From)    : %s (%s)\n", inet_ntoa( addrs), resolve_host_name( iph.saddr));

 addrs.s_addr = iph.daddr;
 printf( "IP Destination address      : %s (%s)\n", inet_ntoa( addrs), resolve_host_name( iph.daddr));

}

void print_header_ipoptions( char *buf, int size)
{
 printf( "IP Options                  :\n");
 print_ala_hexdump( buf, size);
}

void print_header_tcp( struct tcphdr tcph)
{
 printf( "******* TCP Header *******\n");

 printf( "Source port address (From)  : %d/%s\n", htons( tcph.th_sport), get_serv_name( tcph.th_sport));
 printf( "Destination port address    : %d/%s\n", htons( tcph.th_dport), get_serv_name( tcph.th_dport));
 
 printf( "Sequence Number             : %lu (0x%lx)\n", htonl(tcph.th_seq), htonl(tcph.th_seq));
 printf( "Acknowledgement Number      : %lu (0x%lx)\n", htonl(tcph.th_ack), htonl(tcph.th_ack));
 printf( "TCP Header Length           : %-5d (0x%x)\n" , tcph.th_off       , tcph.th_off       );
 printf( "Reserved                    : %-5d (0x%x)\n" , tcph.th_x2        , tcph.th_x2        );
 printf( "FIN flag                    : %s\n", tcph.th_flags & TH_FIN  ? "ON" : "OFF");
 printf( "SYN flag                    : %s\n", tcph.th_flags & TH_SYN  ? "ON" : "OFF");
 printf( "RST flag                    : %s\n", tcph.th_flags & TH_RST  ? "ON" : "OFF");
 printf( "PUSH flag                   : %s\n", tcph.th_flags & TH_PUSH ? "ON" : "OFF");
 printf( "ACK flag                    : %s\n", tcph.th_flags & TH_ACK  ? "ON" : "OFF");
 printf( "URG flag                    : %s\n", tcph.th_flags & TH_URG  ? "ON" : "OFF");
 printf( "Window size                 : %-5d (0x%x)\n", htons( tcph.th_win), htons( tcph.th_win));
 printf( "TCP checksum                : %-5d (0x%x)\n", htons( tcph.th_sum), htons( tcph.th_sum));
 printf( "Urgent pointer              : %-5d (0x%x)\n", htons( tcph.th_urp), htons( tcph.th_urp));

}

void print_header_udp( struct udphdr udph)
{
 printf( "******* UDP Header *******\n");
 printf( "Source port address (From)  : %d/%s\n"      , htons( udph.source), get_serv_name( udph.source));
 printf( "Destination port address    : %d/%s\n"      , htons( udph.dest  ), get_serv_name( udph.dest)  );
 printf( "UDP Length                  : %-5d (0x%x)\n", htons( udph.len   ), htons( udph.len)           );
 printf( "UDP Checksum                : %-5d (0x%x)\n", htons( udph.check ), htons( udph.check)         );

}

void print_header_icmp( struct icmphdr ich)
{
 struct in_addr addrs;
 char type[ 1024];
 char code[ 1024];
 char temp[ 1024];
 int gateway = 0;
 
 printf( "******* ICMP Header *******\n");

 sprintf( code, "%-5d", ich.code);

 switch( ich.type) {
         case ICMP_ECHOREPLY     : strcpy( type, "Echo Reply (ICMP_ECHOREPLY)");
                                   break;
         case ICMP_DEST_UNREACH  : strcpy( type, "Destination Unreachable (ICMP_DEST_UNREACH)");
                                   switch( ich.code) {
                                           case ICMP_NET_UNREACH   : strcpy( code, "Destination Net Unreachable (ICMP_NET_UNREACH)"      ); break;
                                           case ICMP_HOST_UNREACH  : strcpy( code, "Destination Host Unreachable (ICMP_HOST_UNREACH)"    ); break;
                                           case ICMP_PROT_UNREACH  : strcpy( code, "Destination Protocol Unreachable (ICMP_PROT_UNREACH)"); break;
                                           case ICMP_PORT_UNREACH  : strcpy( code, "Destination Port Unreachable (ICMP_PORT_UNREACH)"    ); break;
                                           case ICMP_FRAG_NEEDED   : strcpy( code, "Fragmentation needed and DF set (ICMP_FRAG_NEEDED)"  ); break;
                                           case ICMP_SR_FAILED     : strcpy( code, "Source Route Failed (ICMP_SR_FAILED)"                ); break;
                                           case ICMP_NET_UNKNOWN   : strcpy( code, "???? (ICMP_NET_UNKNOWN)"                             ); break;
                                           case ICMP_HOST_UNKNOWN  : strcpy( code, "???? (ICMP_HOST_UNKNOWN)"                            ); break;
                                           case ICMP_HOST_ISOLATED : strcpy( code, "???? (ICMP_HOST_ISOLATED)"                           ); break;
                                           case ICMP_NET_ANO       : strcpy( code, "???? (ICMP_NET_ANO)"                                 ); break;
                                           case ICMP_HOST_ANO      : strcpy( code, "???? (ICMP_HOST_ANO)"                                ); break;
                                           case ICMP_NET_UNR_TOS   : strcpy( code, "???? (ICMP_NET_UNR_TOS)"                             ); break;
                                           case ICMP_HOST_UNR_TOS  : strcpy( code, "???? (ICMP_HOST_UNR_TOS)"                            ); break;
                                           case ICMP_PKT_FILTERED  : strcpy( code, "Packet filtered (ICMP_PKT_FILTERED)"                 ); break;
                                           case ICMP_PREC_VIOLATION: strcpy( code, "Precedence violation (ICMP_PREC_VIOLATION)"          ); break;
                                           case ICMP_PREC_CUTOFF   : strcpy( code, "Precedence cut off (ICMP_PREC_CUTOFF)"               ); break;
                                           default                 : sprintf( code, "Dest Unreachable, Bad Code: %d", ich.code           ); break;
                                   }
                                   break;
         case ICMP_SOURCE_QUENCH : strcpy( type, "Source Quench (ICMP_SOURCE_QUENCH)");
                                   break;
         case ICMP_REDIRECT      : strcpy( type, "Redirect (change route) (ICMP_REDIRECT)");
                                   switch( ich.code) {
                                           case ICMP_REDIR_NET    : strcpy( temp, "Redirect Network (ICMP_REDIR_NET)"                       ); break;
                                           case ICMP_REDIR_HOST   : strcpy( temp, "Redirect Host (ICMP_REDIR_HOST)"                         ); break;
                                           case ICMP_REDIR_NETTOS : strcpy( temp, "Redirect Type of Service and Network (ICMP_REDIR_NETTOS)"); break;
                                           case ICMP_REDIR_HOSTTOS: strcpy( temp, "Redirect Type of Service and Host (ICMP_REDIR_HOSTTOS)"  ); break;
                                           default                : sprintf( temp, "Redirect, Bad Code: %d", ich.code                       ); break;
                                   }
                                   sprintf( code ,"%s (New addr: 0x%08x)\n", temp, ich.un.gateway);
                                   gateway = 1;
                                   break;
         case ICMP_ECHO          : strcpy( type, "Echo Request (ICMP_ECHO)"          );
                                   break;
         /* See RFC 1256 */
         case 9                  : strcpy( type, "ICMP Router Advertisement Message" );
                                   break;
         case 10                 : strcpy( type, "ICMP Router Solicitation Message"  );
                                   break;
         case ICMP_TIME_EXCEEDED : strcpy( type, "Time Exceeded (ICMP_TIME_EXCEEDED)");
                                   switch( ich.code) {
                                           case ICMP_EXC_TTL     : strcpy( code, "Time to live exceeded (ICMP_EXC_TTL)"             ); break;
                                           case ICMP_EXC_FRAGTIME: strcpy( code, "Frag reassembly time exceeded (ICMP_EXC_FRAGTIME)"); break;
                                           default               : sprintf( code, "Time exceeded, Bad Code: %d", ich.code           ); break;
                                   }
                                   break;
         case ICMP_PARAMETERPROB : sprintf( type, "Parameter problem: pointer = 0x%02x (ICMP_PARAMETERPROB)", ich.un.gateway);
                                   break;
         case ICMP_TIMESTAMP     : strcpy( type, "Timestamp Request (ICMP_TIMESTAMP)"     );
                                   break;
         case ICMP_TIMESTAMPREPLY: strcpy( type, "Timestamp Reply (ICMP_TIMESTAMPREPLY)"  );
                                   break;
         case ICMP_INFO_REQUEST  : strcpy( type, "Information Request (ICMP_INFO_REQUEST)");
                                   break;
         case ICMP_INFO_REPLY    : strcpy( type, "Information Reply (ICMP_INFO_REPLY)"    );
                                   break;
         case ICMP_ADDRESS       : strcpy( type, "Address Mask Request (ICMP_ADDRESS)"    );
                                   break;
         case ICMP_ADDRESSREPLY  : strcpy( type, "Address Mask Reply (ICMP_ADDRESSREPLY)" );
                                   break;
         /* See RFC 1788 */
         case 37                 : strcpy( type, "Domain Name Request"                    );
                                   break;
         case 38                 : strcpy( type, "Domain Name Reply"                      );
                                   break;
         default                 : sprintf( type, "Bad ICMP type: %d", ich.type           );
                                   break;
    } /* end switch( ich.type) */

 printf( "Type                        : %s (%d)\n", type, ich.type);
 printf( "Code                        : %s (%d)\n", code, ich.code);

 if( gateway) {
     addrs.s_addr = ich.un.gateway;
     printf( "Gateway                     : %s (%s)\n"  , inet_ntoa( addrs)           , resolve_host_name( ich.un.gateway));
 } else {
     printf( "Identification              : %-5d (0x%x)\n", htons( ich.un.echo.id)      , htons( ich.un.echo.id)            );
     printf( "Sequence                    : %-5d (0x%x)\n", htons( ich.un.echo.sequence), htons( ich.un.echo.sequence)      );
 }

 printf( "CheckSum                    : %-5d (0x%x)\n"    , htons( ich.checksum)        , htons( ich.checksum)              );

}

void print_data( char *buf, char *pkttype, int size)
{
 printf( "******* %s Data *******\n", pkttype);

 print_ala_hexdump( buf, size);
}

void print_ala_hexdump( char *buf, int size)
{
 int hsize = 0; /* Hex size */
 int pos   = 0;
 
 for( pos = 0; pos < size; pos += 16) {
        printf( "%07X  ", pos);

        hsize = (size - pos) > 16 ? 16: size - pos;

        print_hex(   buf + pos, hsize);
        print_ascii( buf + pos, hsize);
        printf( "\n");
 }

}

void print_hex( char *buf, int size)
{
 int j = 0, sp = 1;

 while( j < size) {
        printf( "%02X", (unsigned char) buf[ j]);
        j++;
        if( sp == 4) {
             printf( "  ");
             sp = 1;
        } else {
             printf( " ");
             sp++;
        }
 }

 while( j++ < 16) {
        printf( "  ");
        if( sp == 4) {
             printf( "  ");
             sp = 1;
        } else {
             printf( " ");
             sp++;
        }
 }

}

void print_ascii( char *buf, int size)
{
 int i;

 for( i = 0; i < size; i++)
      printf( "%c", buf[ i] > 31 && buf[ i] < 127 ? buf[ i] : '.');

 while( i++ < 16)
        printf( " ");

}

